% FEEC/Unicamp
% 28/05/2017
% Unit-growing learning for one-hidden layer neural networks
% Hidden nodes with different nonlinearities and linear output nodes with bias
% Maximum order of the Hermite polynomials is defined arbitrarily
% << With backfitting >>
%
% Restrictions (can be eliminated if necessary):
% 1) number of inputs has to be less than (or equal to) 7
%
% Adjustable parameters:
% nMAX: maximum number of neurons that can be used
% P: order of the Hermite polynomials
% j: When j=0, then the training will start. When j>0, then the training
%    will continue introducing neuron j+1. The other j neurons should have
%    been defined in a previous simulation or by the introduction of
%    initial knowledge.
% threshold1: stopping criterion (mean square error)
% threshold2: stopping criterion (evolution of the adjustment in each neuron)
% threshold3: stopping criterion (evolution of the adjustment in each set of
%          already introduced neurons)
% threshold4(calcv.m): stopping criterion for 2nd order optimization (gradient norm)
% threshold5(calcv.m): minimum value for the first order optimization step (alpha)
%
% Auxiliary functions: m_norm.m / calcbias.m / calcu.m / calcvi.m / calcc.m / itervf.m
%                      factiv.m / update_node.m / retrosp.m
clear all;
nMAX = input('Insert the maximum number of hidden neurons (Suggestion: 2 to 5): ');
P = input('Insert the maximum degree of the Hermite polynomials (Suggestion: 9 to 12): ');
j = 0; % If j > 0, the training is restarted to introduce additional neurons
threshold1 = 5.0e-3; % delta (mean square error)
threshold2 = 1.0e-4; % epsilon (evolution of the adjustment in each neuron)
threshold3 = 1.0e-3; % gamma (evolution of the adjustment in each set of neurons)
load data; % data.mat should contain X and S
D = S;
[N,m] = size(X);disp(sprintf('Number of input-output data=%d',N));
r = length(S(1,:));
if j > 0,
	load vv;load vw;load vc;load vbias;
	Srn = zeros(N,r);
	for a=1:j,
		Srn = Srn + factiv(X*v(:,a),c(a,:))*w(a,:);
		for k=1:r,
			Srn(:,k) = Srn(:,k)+bias(a,k);
		end
	end
	D = S-Srn;
end
E = m_norm(D);disp(sprintf('RMSE=%g/SC=%g',E,threshold1)); % SC -> stopping criterion
Eold = E;Enew = 0;
dir_init_chosen = [];
v0 = [];
j = j+1;
while (E > threshold1) & (j <= nMAX), % training will introduce at most nMAX neurons
	disp(sprintf('Neuron %d',j));
	[bias(j,:),D] = calcbias(D);
	u = calcu(D);
	vi = calcvi(m);
	evalv = [];
	sensor = 0;
	for qq=1:length(vi(1,:)),
		if sensor == 0,
			caux(qq,:) = calcc(X*vi(:,qq),u,P);
			[vi(:,qq),caux(qq,:)] = itervf(vi(:,qq),caux(qq,:),u,X,threshold2,qq);
			y = factiv(X*vi(:,qq),caux(qq,:));
			evalv(qq) = (y'*u)^2/(y'*y);
		else
			evalv(qq) = 0;
		end
		if evalv(qq) > 0.9,
			sensor = 1;
		end
	end
	[qq,pos] = max(evalv);dir_init_chosen = [dir_init_chosen;pos];v0 = [v0 vi(:,pos)];
	v(:,j) = vi(:,pos);disp('v chosen');disp(v(:,j)');
	c(j,:) = caux(pos,:);
	[v(:,j),c(j,:),w(j,:),Sj] = update_node(v(:,j),c(j,:),u,X,D,threshold2,j);
	D = D-Sj;
%	disp('v');disp(v(:,j)');disp('c');disp(c(j,:));disp('w');disp(w(j,:));
	Enew = m_norm(D);Eold = 2*Enew;
	cont = 1;
	while (abs(Eold-Enew)/Eold > threshold3),
		disp(sprintf('cont=%d',cont));
		disp(sprintf('%g/%g',abs(Eold-Enew)/Eold,threshold3));
		for a=1:j,
			Srn = zeros(N,r);
			for b=1:j,
				if b ~= a,
					Srn = Srn + factiv(X*v(:,b),c(b,:))*w(b,:);
					for k=1:r,
						Srn(:,k) = Srn(:,k)+bias(b,k);
					end
				end
			end
			D = S-Srn;
			[bias(a,:),D] = calcbias(D);
			u = calcu(D);
			[v(:,a),c(a,:),w(a,:),Sj] = update_node(v(:,a),c(a,:),u,X,D,threshold2,a);
		end
		for k=1:r,
			Sj(:,k) = Sj(:,k)+bias(j,k);
		end
		D = S-(Srn+Sj);
		Eold = Enew;Enew = m_norm(D);
		cont = cont+1;
	end
	j = j+1;
	E = Enew;disp(sprintf('RMSE=%g/SC=%g',E,threshold1));
	retrosp(X,S,w,v,c,bias,E,dir_init_chosen,v0,j);
end
disp('RMSE');disp(m_norm(D));
disp('w');disp(w);disp('v^T');disp(v');disp('c^T');disp(c');disp('bias^T');disp(bias');
save vw w;save vv v;save vc c;save vbias bias;
